package com.engage.zapier.service.impl;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.engage.zapier.constant.ServiceConstant;
import com.engage.zapier.entity.common.CRMFieldEntity;
import com.engage.zapier.entity.common.LeadEntity;
import com.engage.zapier.entity.common.ModuleEntity;
import com.engage.zapier.entity.common.UserEntity;
import com.engage.zapier.entity.master.ZapierWebhookDetailsEntity;
import com.engage.zapier.exception.ZapierServiceException;
import com.engage.zapier.repository.common.CRMFieldRepository;
import com.engage.zapier.repository.common.LeadRepository;
import com.engage.zapier.repository.common.ModuleRepository;
import com.engage.zapier.repository.common.UserRepository;
import com.engage.zapier.repository.master.ZapierWebhookDetailsRepository;
import com.engage.zapier.request.WebhookRequest;
import com.engage.zapier.response.CrmFieldResponse;
import com.engage.zapier.response.LeadResponse;
import com.engage.zapier.response.PicklistResponse;
import com.engage.zapier.response.ServiceResponse;
import com.engage.zapier.response.UserResponse;
import com.engage.zapier.response.ZapierWebhookResponse;
import com.engage.zapier.service.ZapierService;
import com.engage.zapier.utility.MasterDataForStandardFields;

@Service
public class ZapierServiceImpl implements ZapierService {

	private static final Logger LOGGER = LoggerFactory.getLogger(ZapierServiceImpl.class);

	@Autowired
	private UserRepository userRepository;

	@Autowired
	private ModuleRepository moduleRepository;

	@Autowired
	private CRMFieldRepository crmFieldRepository;

	@Autowired
	private MasterDataForStandardFields masterDataForStandardFields;

	@Autowired
	private ZapierWebhookDetailsRepository zapierWebhookDetailsRepository;

	@Autowired
	private LeadRepository leadRepository;

	@SuppressWarnings("unchecked")
	@Override
	public <T> ServiceResponse<T> getUserInfo(UserEntity userEntity) {
		UserResponse userResponse = new UserResponse();
		userResponse.setFirstName(userEntity.getFirstName());
		userResponse.setLastName(userEntity.getLastName());
		userResponse.setMobile(userEntity.getMobile());
		userResponse.setPhone(userEntity.getPhoneNumber());
		userResponse.setFullName(userEntity.getUserFullName());
		userResponse.setEmail(
				userEntity.getUserMappingsByUserMappingId().getUserAuthenticationByAuthenticationId().getEmailId());
		return new ServiceResponse<T>((T) userResponse, HttpStatus.OK);
	}

	@SuppressWarnings("unchecked")
	@Override
	@Transactional(readOnly = true)
	public <T> ServiceResponse<T> getLeadFields(UserEntity userEntity) {
		ModuleEntity moduleEntity = null;
		T result = null;
		Long orgId = userEntity.getUserMappingsByUserMappingId().getSubscriberOrganizationByOrganizationId()
				.getOrganizationId();
		try {
			moduleEntity = moduleRepository.findByModuleNameAndIsDeleted(ServiceConstant.LEAD_MODULE, Boolean.FALSE);
			if (Optional.ofNullable(moduleEntity).isEmpty()) {
				throw new ZapierServiceException(HttpStatus.NO_CONTENT.value(), "Lead module not found.");
			}
			List<CRMFieldEntity> crmFieldEntities = getCrmFields.apply(moduleEntity, orgId);
			if (CollectionUtils.isNotEmpty(crmFieldEntities)) {
				result = (T) generateCrmFieldResponse.apply(crmFieldEntities, orgId);
			} else {
				throw new ZapierServiceException(HttpStatus.NO_CONTENT.value(),
						"Crm fields not available for Lead module.");
			}
		} catch (Exception e) {
			LOGGER.error("Exception occured while getting crm fields::{}", e.getMessage());
			throw new ZapierServiceException(HttpStatus.INTERNAL_SERVER_ERROR.value(), e.getMessage());
		}
		return new ServiceResponse<T>((T) result, HttpStatus.OK);
	}

	private BiFunction<ModuleEntity, Long, List<CRMFieldEntity>> getCrmFields = (moduleEntity,
			orgId) -> crmFieldRepository.findCRMFieldListByModule(Arrays.asList(orgId, 0L),
					moduleRepository.getModuleListByParentId(moduleEntity));

	private Function<Long, List<PicklistResponse>> getAllUser = orgId -> {
		List<PicklistResponse> response = null;
		List<UserEntity> userEntities = userRepository.getActiveUsersEntityOfOrganization(orgId, Boolean.FALSE,
				Boolean.TRUE, Boolean.FALSE);
		if (CollectionUtils.isNotEmpty(userEntities)) {
			return userEntities.stream().map(user -> {
				return new PicklistResponse(user.getUserId(), user.getUserFullName());
			}).collect(Collectors.toList());
		}
		return response;
	};

	private BiFunction<List<CRMFieldEntity>, Long, List<CrmFieldResponse>> generateCrmFieldResponse = (crmFieldEntities,
			orgId) -> {
		List<CrmFieldResponse> response = new ArrayList<>();
		crmFieldEntities.parallelStream().forEach(crmField -> {
			CrmFieldResponse crmFieldResponse = new CrmFieldResponse();
			crmFieldResponse.setSection(crmField.getModuleId().getModuleName());
			crmFieldResponse.setSectionId(crmField.getModuleId().getModuleId());
			if (StringUtils.isNoneBlank(crmField.getDefaultValue()))
				crmFieldResponse.setDefaultValue(crmField.getDefaultValue());
			crmFieldResponse.setId(crmField.getCrmFieldId());
			if (StringUtils.isNoneBlank(crmField.getFieldLength()))
				crmFieldResponse.setFieldLength(crmField.getFieldLength());
			crmFieldResponse.setIsMandatory(crmField.getIsMandatory());
			crmFieldResponse.setSortOrder(crmField.getSortOrder());
			if (StringUtils.isNoneBlank(crmField.getFieldName()))
				crmFieldResponse.setFieldName(crmField.getFieldName());
			crmFieldResponse.setFieldType(crmField.getMasterDataByFieldType().getDataValue());
			if (crmField.getIsCustom() == 0) {
				crmFieldResponse.setFormFieldName(crmField.getFormFieldName());
				if (StringUtils.equalsIgnoreCase(ServiceConstant.PICKLIST_TYPE,
						crmField.getMasterDataByFieldType().getDataValue())
						&& !StringUtils.equalsIgnoreCase(ServiceConstant.CRM_FIELD_LEAD_OWNER,
								crmField.getFormFieldName())
						&& !StringUtils.equalsIgnoreCase(ServiceConstant.CRM_FIELD_STATE, crmField.getFormFieldName())
						&& !StringUtils.equalsIgnoreCase(ServiceConstant.CRM_FIELD_COUNTRY,
								crmField.getFormFieldName())
						&& !StringUtils.equalsIgnoreCase(ServiceConstant.CRM_FIELD_PRODUCT_INTEREST,
								crmField.getFormFieldName())) {
					crmFieldResponse.setPickListValues(
							masterDataForStandardFields.getMasterDataListByFormFieldName(crmField.getFormFieldName()));
				} else if (StringUtils.equalsIgnoreCase(ServiceConstant.CRM_FIELD_LEAD_OWNER,
						crmField.getFormFieldName())) {
					crmFieldResponse.setPickListValues(getAllUser.apply(orgId));
				}
			} else if (crmField.getIsCustom() == 1
					&& CollectionUtils.isNotEmpty(crmField.getPickListValueEntityByCrmFieldId())) {
				crmFieldResponse
						.setPickListValues(crmField.getPickListValueEntityByCrmFieldId().stream().map(pickList -> {
							return new PicklistResponse(pickList.getPickListId(), pickList.getPickListValue());
						}).collect(Collectors.toList()));
			}
			response.add(crmFieldResponse);
		});
		List<CrmFieldResponse> sortedResponse = response.stream()
				.sorted(Comparator.comparingLong(CrmFieldResponse::getSectionId)).collect(Collectors.toList());
		return sortedResponse;
	};

	@SuppressWarnings("unchecked")
	@Override
	@Transactional(readOnly = false)
	public <T> ServiceResponse<T> subscribeWebHook(WebhookRequest webhookRequest, UserEntity userEntity, String dsmId) {
		ZapierWebhookDetailsEntity webhookDetailsEntity = new ZapierWebhookDetailsEntity();
		if (StringUtils.isNoneBlank(webhookRequest.getEvent()) && StringUtils.isNoneBlank(webhookRequest.getUrl())) {
			webhookDetailsEntity.setDsmId(dsmId);
			webhookDetailsEntity.setCreatedDate(LocalDateTime.now());
			webhookDetailsEntity.setEvent(webhookRequest.getEvent());
			webhookDetailsEntity.setUrl(webhookRequest.getUrl());
			webhookDetailsEntity.setOrganizationId(userEntity.getUserMappingsByUserMappingId()
					.getSubscriberOrganizationByOrganizationId().getOrganizationId());
			zapierWebhookDetailsRepository.save(webhookDetailsEntity);
			return new ServiceResponse<T>((T) generateWebHookResponse.apply(webhookDetailsEntity), HttpStatus.OK);
		} else {
			throw new ZapierServiceException(HttpStatus.NOT_ACCEPTABLE.value(), "Event and URL can not be blank.");
		}
	}

	private Function<ZapierWebhookDetailsEntity, ZapierWebhookResponse> generateWebHookResponse = entity -> new ZapierWebhookResponse(
			entity.getId(), entity.getUrl(), entity.getEvent());

	@Override
	@Transactional(readOnly = false)
	public <T> ServiceResponse<T> unsubscribeWebHook(Long id, UserEntity userEntity,
			String dsmId) {
		Optional<ZapierWebhookDetailsEntity> zapierWebhookDetailsEntityOptional = zapierWebhookDetailsRepository
				.findById(id);
		if (zapierWebhookDetailsEntityOptional.isEmpty()) {
			throw new ZapierServiceException(HttpStatus.NOT_ACCEPTABLE.value(), "Please provide valid subscriber id.");
		}
		zapierWebhookDetailsRepository.delete(zapierWebhookDetailsEntityOptional.get());
		return new ServiceResponse<T>(Boolean.TRUE, HttpStatus.OK);
	}

	@SuppressWarnings("unchecked")
	@Override
	@Transactional(readOnly = true)
	public <T> ServiceResponse<T> getLeadList(UserEntity userEntity) {
		try {
			Page<LeadEntity> entities = leadRepository
					.findAllByLeadOwnerInAndIsDeleted(
							userRepository.getActiveUsersEntityOfOrganization(
									userEntity.getUserMappingsByUserMappingId()
											.getSubscriberOrganizationByOrganizationId().getOrganizationId(),
									Boolean.FALSE, Boolean.TRUE, Boolean.FALSE),Boolean.TRUE,
							PageRequest.of(0, 10));

			if (!entities.hasContent()) {
				throw new ZapierServiceException(HttpStatus.NO_CONTENT.value(), "Lead not available.");
			}
			return new ServiceResponse<T>((T) generateLeadResponse.apply(entities.getContent()), HttpStatus.OK);
		} catch (Exception e) {
			LOGGER.error("Exception occured while executing getLeadList :: {}", e.getMessage());
			throw new ZapierServiceException(HttpStatus.INTERNAL_SERVER_ERROR.value(), e.getMessage());
		}
	}

	private Function<List<LeadEntity>, List<LeadResponse>> generateLeadResponse = leadEntities -> {
		List<LeadResponse> leadResponse = new ArrayList<>();
		leadEntities.forEach(lead->{
			leadResponse.add(new LeadResponse(lead));
		});
		return leadResponse;
	};
	

}
